/*
 * COPYTAPE.C
 *
 * This program duplicates magnetic tapes, preserving the
 * blocking structure and placement of tape marks.
 *
 * This program was updated at
 *
 *	U.S. Army Artificial Intelligence Center
 *	HQDA (Attn: DACS-DMA)
 *	Pentagon
 *	Washington, DC  20310-0200
 *
 *	Phone: (202) 694-6900
 *
 **************************************************
 *
 *	THIS PROGRAM IS IN THE PUBLIC DOMAIN
 *
 **************************************************
 *
 * July 1986		David S. Hayes
 *		Made data file format human-readable.
 *
 * April 1985		David S. Hayes
 *		Original Version.
 */


#include <stdio.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mtio.h>
#include <sys/file.h>

extern int      errno;

#define BUFLEN		64*1024	/* max tape block size */
#define TAPE_MARK	-100	/* return record length if we read a
				 * tape mark */
#define END_OF_TAPE	-101	/* 2 consecutive tape marks */
#define FORMAT_ERROR	-102	/* data file munged */

int             totape = 0,	/* treat destination as a tape drive */
                fromtape = 0;	/* treat source as a tape drive */

int             Debug = 0, Check = 0,
                Stdin = 0, Stdout = 0,
                verbose = 0;	/* tell what we're up to */

char           *source = "stdin",
               *dest = "stdout";

char            inbuf[BUFLEN],
                outbuf[BUFLEN],
                tapebuf[BUFLEN];

main(argc, argv)
    int             argc;
    char           *argv[];
{
    FILE           *from = stdin,
                   *to = stdout;/* our input and output */
    int             len;	/* number of bytes in record */
    int             skip = 0;	/* number of files to skip before
				 * copying */
    unsigned int    limit = 0xffffffff;
    int             i;
    struct mtget    status;


    for (i = 1; i < argc && argv[i][0] == '-' && strlen(argv[i]) != 1; i++) {
	switch (argv[i][1]) {
	  case 's':		/* skip option */
	    skip = atoi(&argv[i][2]);
	    break;

	  case 'l':
	    limit = atoi(&argv[i][2]);
	    break;

	  case 'f':		/* from tape option */
	    fromtape = 1;
	    break;
#ifdef foo
	  case 'i':		/* use stdin anyway */
	    Stdin = 1;
	    break; 

	  case 'o':		/* use stdout anyway */
	    Stdout = 1;
	    break;
#endif
	  case 't':		/* to tape option */
	    totape = 1;
	    break;

	  case 'v':		/* be wordy */
	    verbose = 1;
	    break;

	  case 'd':		/* be wordier */
	    ++Debug;
	    break;

	  case 'c':		/* check */
	    Check = 1;
	    break;

	  default:
	    fprintf(stderr, "usage: copytape [-f] [-t] [-lnn] [-snn] [-v] from to\n");
	    exit(-1);
	}
    }

    if (i < argc && strcmp(argv[i], "-") == 0) {  /* "-" == (source is stdin) */
        Stdin = 1;
        i++;
    }
    else if (i < argc) {                     /* source not stdin */
	from = freopen(argv[i], "r", stdin);
	source = argv[i];
	if (from == NULL) {
	    perror("copytape: input open failed");
	    exit(-1);
	}
	i++;
    }

    if (i < argc && strcmp(argv[i], "-") == 0) {  /* "-" == (dest is stdout) */
        Stdout = 1;
        i++;
    }
    else if (i < argc) {                    /* dest not stdin */
	to = freopen(argv[i], "w", stdout);
	dest = argv[i];
	if (to == NULL) {
	    perror("copytape: output open failed");
	    exit(-1);
	}
	i++;
    }
    if (i < argc)
	perror("copytape: extra arguments ignored");

    /*
     * Determine if source and/or destination is a tape device. Try to
     * issue a magtape ioctl to it.  If it doesn't error, then it was a
     * magtape. 
     */

    errno = 0;
    ioctl(fileno(from), MTIOCGET, &status);
    fromtape |= errno == 0;
    errno = 0;
    ioctl(fileno(to), MTIOCGET, &status);
    totape |= errno == 0;
    errno = 0;

    if (verbose) {
	fprintf(stderr, "copytape: from %s (%s)\n",
		source, fromtape ? "tape" : "data");
	fprintf(stderr, "          to %s (%s)\n",
		dest, totape ? "tape" : "data");
    }
    /* Establish buffering for our IO. */

    if (fromtape)
	setbuffer(from, inbuf, BUFLEN);
    if (totape)
	setbuffer(to, outbuf, BUFLEN);


    /*
     * Skip number of files, specified by -snnn, given on the command
     * line. This is used to copy second and subsequent files on the
     * tape. 
     */

    if (verbose && skip) {
	fprintf(stderr, "copytape: skipping %d input files\n", skip);
    }
    for (i = 0; i < skip; i++) {
	do {
	    len = input(from);
	} while (len != TAPE_MARK && len != END_OF_TAPE);
	if (len == END_OF_TAPE) {
	    fprintf(stderr, "copytape: only %d files in input\n", i);
	    exit(-1);
	}
    }

    /*
     * Do the copy. 
     */

    i = 0;
    while (limit && len != END_OF_TAPE) {
	do {
	    do {
		len = input(from);
		if (len == FORMAT_ERROR)
		{
		    perror("copytape: data format error - block ignored");
		    exit(1);
		}
	    } while (len == FORMAT_ERROR);

	    if (!Check) output(to, len);

	    if (verbose) {
		switch (len) {
		  case TAPE_MARK:
		    fprintf(stderr, "  copied TAPE_MARK\n");
		    break;

		  case END_OF_TAPE:
		    fprintf(stderr, "  copied END_OF_TAPE\n");
		    break;

		  default:
		    ++i;
		    if (Debug) fprintf(stderr,
				       "\ncopied %d bytes:%d\n", len,i);
		}
	    }
	} while (len > 0);
	limit--;
    }
    exit(0);
}


/*
 * Input up to BUFLEN from a file or tape. If input file is a tape, then
 * do markcount stuff.  Input record length will be supplied by the
 * operating system. 
 *
 * If input file is not a tape, then read an integer from the input.  This
 * integer will specify the number of bytes that comprise the following
 * record, not counting the length of the integer itself. 
 *
 * An input length of zero corresponds to a tape mark. We return the
 * number of bytes read.  If we hit logical EOT (two tape marks in a
 * row), then return -1. 
 */

input(fp)
    FILE           *fp;
{
    static          markcount = 0;	/* number of consecutive tape
					 * marks */
    int             len,
                    l2,
                    nlen;
    char            header[80],
                    flag[20],
                    line[80];

    if (fromtape) {
	len = read(fileno(fp), tapebuf, BUFLEN);
	if (len == 0) {
	    if (++markcount == 2)
		return END_OF_TAPE;
	    else
		return TAPE_MARK;
	}
	markcount = 0;		/* reset tape mark count */
	return len;
    }

    fgets(line, 80, fp);
    if (strlen(line) == 1) fgets(line, 80, fp);
    if (Debug == 2) fprintf(stderr, "\nHeader was:%s",line);
    
    l2 = sscanf(line, "%4s:%s %d", flag, header, &nlen);
    if (l2 < 2 || strcmp(flag, "CPTP") != 0)
    {
        fprintf(stderr, "Bad header looking for CPTP, Was: %s\n", flag);
        fprintf(stderr, "Previous block was:\n");
	fwrite(tapebuf, 1, len, stderr);
	return FORMAT_ERROR;
    }

    if (strcmp(header, "BLOCK") == 0) {
/*	fscanf(fp, "%d\n", &len);	/* how long is the block? */
        len = nlen;
	l2 = fread(tapebuf, sizeof(char), len, fp);
	if (l2 != len)		/* did we get it all? */
	{
	    fprintf(stderr, "Bad len, was: %d vs. %d\n", l2, len);
	    return FORMAT_ERROR;
        }
	if (Check) fwrite(tapebuf, 1, l2, stderr);
    } else if (strcmp(header, "TAPE_MARK") == 0)
	return TAPE_MARK;
    else if (strcmp(header, "END_OF_TAPE") == 0)
	return END_OF_TAPE;
    else 
    {
      fprintf(stderr, "Bad header was: %s\n", header);
      return FORMAT_ERROR;
    }

    return len;
}


/*
 * Copy a buffer out to a file or tape. 
 *
 * If output is a tape, write the record.  A length of zero indicates that
 * a tapemark should be written. 
 *
 * If not a tape, write len to the output file, then the buffer.  
 */

output(fp, len)
    FILE           *fp;
    int             len;
{
    struct mtop     op;

    if (totape && (len == TAPE_MARK || len == END_OF_TAPE)) {
	op.mt_op = MTWEOF;
	op.mt_count = 1;
	ioctl(fileno(fp), MTIOCTOP, &op);
	return;
    }
    if (!totape) {
	switch (len) {
	  case TAPE_MARK:
	    fprintf(fp, "CPTP:TAPE_MARK\n");
	    break;

	  case END_OF_TAPE:
	    fprintf(fp, "CPTP:END_OF_TAPE\n");
	    break;

	  case FORMAT_ERROR:
	    break;

	  default:
	    fprintf(fp, "CPTP:BLOCK %05d\n", len);
	    fwrite(tapebuf, sizeof(char), len, fp);
	    putc('\n', fp);
	}
    } else
	write(fileno(fp), tapebuf, len);
}
